#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "snes_rom_header.h"

//#define DEBUG

static int find_max(int *table, int entries);

int find_and_read_header_from_file(FILE *fptr, struct snes_rom_header *header)
{		
		int res, i; 
		int rom_scores[4] = { 0, 0, 0, 0 };			
		int offsets[4] = { 65984, 65472, 33216, 32704 };
	
		for (i=0; i<4; i++)
		{
			fseek(fptr, offsets[i], SEEK_SET);
			res = read_header_from_file(fptr, header);
			if (res>=0) {
				rom_scores[i] = get_score(header);
#ifdef DEBUG
				printf("Offset %d scores %d\n", offsets[i], rom_scores[i]);
#endif
			}
		}
		

		if ((rom_scores[0]+rom_scores[1]+rom_scores[2]+rom_scores[3])==0) {
			return -1; // invalid rom
		}

		i = find_max(rom_scores, 4);

		fseek(fptr, offsets[i], SEEK_SET);
		res = read_header_from_file(fptr, header);
		if (res>=0) {
			return 0;
		}
	
		return -1;	
}

static int find_max(int *table, int entries)
{
	int i;
	int max_value=0;
	int max_index=0;
	
	for (i=0; i<entries; i++)
	{
		if (table[i]>max_value) { 
			max_value=table[i]; 
			max_index=i;
		}
	}
	return max_index;
}

int get_score(struct snes_rom_header *header)
{
	int score;

	score = test_checksums(header);
	if (!score) { return 0; } // definitely this is not the right location
	
	if (get_rom_type_string(header->rom_type)!=NULL) { score++; }
	if (get_country_string(header->country)!=NULL) { score++; }
	if (get_license_string(header->license)!=NULL) { score++; }	
	
	return score;
}

int test_checksums(struct snes_rom_header *header)
{
		if ((header->inverse_checksum|header->checksum)==0xFFFF) {
				return 1;
		}
		return 0;
}

int read_header_from_file(FILE *fptr, struct snes_rom_header *header)
{
	int res;
		
	// title
	fread(&header->title, 21, 1, fptr);
	header->title[21]=0;

	// makeup
	res = fread(&header->rom_makeup, 1, 1, fptr);
	if (res!=1) { return -1; }
	
	// type
	res = fread(&header->rom_type, 1, 1, fptr);
	if (res!=1) { return -1; }
	
	// size
	res = fread(&header->rom_size, 1, 1, fptr);
	if (res!=1) { return -1; }
	
	// sram_size
	res = fread(&header->sram_size, 1, 1, fptr);
	if (res!=1) { return -1; }
	
	// country
	res = fread(&header->country, 1, 1, fptr);
	if (res!=1) { return -1; }
	
	// license
	res = fread(&header->license, 1, 1, fptr);
	if (res!=1) { return -1; }
	
	// version
	res = fread(&header->game_version, 1, 1, fptr);
	if (res!=1) { return -1; }
	
	// inverse_checksum
	res = fread(&header->inverse_checksum, 2, 1, fptr);
	if (res!=1) { return -1; }
	
	// checksum
	res = fread(&header->checksum, 2, 1, fptr);
	if (res!=1) { return -1; }
	
	// reset_vector
	res = fread(&header->reset_vector, 2, 1, fptr);
	if (res!=1) { return -1; }
	
	return 0;	
}

char *get_rom_type_string(unsigned char rom_type)
{
		switch(rom_type)
		{
				case 0: return "ROM only";
				case 1: return "ROM and RAM";
				case 2: return "ROM and Save RAM";
				case 3: return "ROM and DSP1 chip";
				case 4: return "ROM, RAM and DSP1 chip";
				case 5: return "ROM, Save RAM and DSP1 chip";
				case 19: return "ROM and Super FX chip";
				case 227: return "ROM, RAM and GameBoy data";
				case 246: return "ROM and DSP2 chip";
		}
		return NULL;
}

char *get_country_string(unsigned char country)
{
		switch(country)
		{
				case 0: return "Japan";
				case 1: return "USA";
				case 2: return "Australia, Europe, Oceania and Asia";
				case 3: return "Sweden";
				case 4: return "Finland";
				case 5: return "Denmark";
				case 6: return "France";
				case 7: return "Holland";
				case 8: return "Spain";
				case 9: return "Germany, Australia and Switzerland";
				case 10: return "Italy";
				case 11: return "Hong Kong and China";
				case 12: return "Indonesia";
				case 13: return "Korea";				
		}
		return NULL;
}

char *get_license_string(unsigned char license)
{
		switch(license)
		{
			case 1: return "Nintendo";
			case 3: return "Imagineer-Zoom";
			case 5: return "Zamuse";
			case 6: return "Falcom";
			case 8: return "Capcom";
			case 9: return "HOT-B";
			case 10: return "Jaleco";
			case 11: return "Coconuts";
			case 12: return "Rage Software";
			case 14: return "Technos";
			case 15: return "Mebio Software";
			case 18: return "Gremlin Graphics";
			case 19: return "Electronic Arts";
			case 21: return "COBRA Team";
			case 22: return "Human/Field";
			case 23: return "KOEI";
			case 24: return "Hudson Soft";
			case 26: return "Yanoman";
			case 28: return "Tecmo";
			case 30: return "Open System";
			case 31: return "Virgin Games";
			case 32: return "KSS";
			case 33: return "Sunsoft";
			case 34: return "POW";
			case 35: return "Micro World";
			case 38: return "Enix";
			case 39: return "Loriciel/Electro Brain";
			case 40: return "Kemco";
			case 41: return "Seta Co.,Ltd.";
			case 45: return "Visit Co.,Ltd.";
			case 49: return "Carrozzeria";
			case 50: return "Dynamic";
			case 51: return "Nintendo";
			case 52: return "Magifact";
			case 53: return "Hect";
			case 60: return "Empire Software";
			case 61: return "Loriciel";
			case 64: return "Seika Corp.";
			case 65: return "UBI Soft";
			case 70: return "System 3";
			case 71: return "Spectrum Holobyte";
			case 73: return "Irem";
			case 75: return "Raya Systems/Sculptured Software";
			case 76: return "Renovation Products";
			case 77: return "Malibu Games/Black Pearl";
			case 79: return "U.S. Gold";
			case 80: return "Absolute Entertainment";
			case 81: return "Acclaim";
			case 82: return "Activision";
			case 83: return "American Sammy";
			case 84: return "GameTek";
			case 85: return "Hi Tech Expressions";
			case 86: return "LJN Toys";
			case 90: return "Mindscape";
			case 93: return "Tradewest";
			case 95: return "American Softworks Corp.";
			case 96: return "Titus";
			case 97: return "Virgin Interactive Entertainment";
			case 98: return "Maxis";
			case 103: return "Ocean";
			case 105: return "Electronic Arts";
			case 107: return "Laser Beam";
			case 110: return "Elite";
			case 111: return "Electro Brain";
			case 112: return "Infogrames";
			case 113: return "Interplay";
			case 114: return "LucasArts";
			case 115: return "Parker Brothers";
			case 117: return "STORM";
			case 120: return "THQ Software";
			case 121: return "Accolade Inc.";
			case 122: return "Triffix Entertainment";
			case 124: return "Microprose";
			case 127: return "Kemco";
			case 128: return "Misawa";
			case 129: return "Teichio";
			case 130: return "Namco Ltd.";
			case 131: return "Lozc";
			case 132: return "Koei";
			case 134: return "Tokuma Shoten Intermedia";
			case 136: return "DATAM-Polystar";
			case 139: return "Bullet-Proof Software";
			case 140: return "Vic Tokai";
			case 142: return "Character Soft";
			case 143: return "I''Max";
			case 144: return "Takara";
			case 145: return "CHUN Soft";
			case 146: return "Video System Co., Ltd.";
			case 147: return "BEC";
			case 149: return "Varie";
			case 151: return "Kaneco";
			case 153: return "Pack in Video";
			case 154: return "Nichibutsu";
			case 155: return "TECMO";
			case 156: return "Imagineer Co.";
			case 160: return "Telenet";
			case 164: return "Konami";
			case 165: return "K.Amyusement Leasing Co.";
			case 167: return "Takara";
			case 169: return "Technos Jap.";
			case 170: return "JVC";
			case 172: return "Toei Animation";
			case 173: return "Toho";
			case 175: return "Namco Ltd.";
			case 177: return "ASCII Co. Activision";
			case 178: return "BanDai America";
			case 180: return "Enix";
			case 182: return "Halken";
			case 186: return "Culture Brain";
			case 187: return "Sunsoft";
			case 188: return "Toshiba EMI";
			case 189: return "Sony Imagesoft";
			case 191: return "Sammy";
			case 192: return "Taito";
			case 194: return "Kemco";
			case 195: return "Square";
			case 196: return "Tokuma Soft";
			case 197: return "Data East";
			case 198: return "Tonkin House";
			case 200: return "KOEI";
			case 202: return "Konami USA";
			case 203: return "NTVIC";
			case 205: return "Meldac";
			case 206: return "Pony Canyon";
			case 207: return "Sotsu Agency/Sunrise";
			case 208: return "Disco/Taito";
			case 209: return "Sofel";
			case 210: return "Quest Corp.";
			case 211: return "Sigma";
			case 214: return "Naxat";
			case 216: return "Capcom Co., Ltd.";
			case 217: return "Banpresto";
			case 218: return "Tomy";
			case 219: return "Acclaim";
			case 221: return "NCS";
			case 222: return "Human Entertainment";
			case 223: return "Altron";
			case 224: return "Jaleco";
			case 226: return "Yutaka";
			case 228: return "T&ESoft";
			case 229: return "EPOCH Co.,Ltd.";
			case 231: return "Athena";
			case 232: return "Asmik";
			case 233: return "Natsume";
			case 234: return "King Records";
			case 235: return "Atlus";
			case 236: return "Sony Music Entertainment";
			case 238: return "IGS";
			case 241: return "Motown Software";
			case 242: return "Left Field Entertainment";
			case 243: return "Beam Software";
			case 244: return "Tec Magik";
			case 249: return "Cybersoft";
			case 255: return "Hudson Soft";			  
		}
		return NULL;
}

